"""Request handler utility classes.

Includes classes for serving templates with common variables and XRD[S] and JRD
files like host-meta and friends.
"""
import calendar
import functools
import logging
import os
import threading
import urllib.parse

import cachetools
from google.cloud import ndb
import jinja2
import webapp2

from . import util
from .util import json_dumps, json_loads

JINJA_ENV = jinja2.Environment(
  loader=jinja2.FileSystemLoader(('.', 'templates')),
  autoescape=True,
)
JINJA_ENV.globals.update({
  'EPOCH': util.EPOCH,
  'timestamp': lambda dt: calendar.timegm(dt.utctimetuple()),
})


def handle_exception(self, e, debug):
  """A webapp2 exception handler that propagates HTTP exceptions into the response.

  Use this as a :meth:`webapp2.RequestHandler.handle_exception()` method by
  adding this line to your handler class definition::

    handle_exception = handlers.handle_exception
  """
  code, body = util.interpret_http_exception(e)
  if code:
    self.response.set_status(int(code))
    self.response.write('Upstream server request failed: %s' % e
                        if code in ('502', '504')
                        else 'HTTP Error %s: %s' % (code, body))
  else:
    raise


def redirect(from_domain, to_domain):
  """:class:`webapp2.RequestHandler` decorator that 301 redirects to a new domain.

  Preserves scheme, path, and query.

  Args:
    from_domain: string or sequence of strings
    to_domain: strings
  """
  if isinstance(from_domain, str):
    from_domain = [from_domain]

  def decorator(method):
    def wrapper(self, *args, **kwargs):
      parts = list(urllib.parse.urlparse(self.request.url))
      # not using self.request.host because it includes port
      if parts[1] in from_domain:  # netloc
        parts[1] = to_domain
        return self.redirect(urllib.parse.urlunparse(parts), permanent=True)
      else:
        return method(self, *args, **kwargs)

    return wrapper

  return decorator


def cache_response(expiration, size=20 * 1000 * 1000):  # 20 MB
  """:class:`webapp2.RequestHandler` method decorator that caches the response in memory.

  Includes a `cache_clear()` function that clears all cached responses.

  Ideally this would be just a thin wrapper around the
  :func:`cachetools.cachedmethod` decorator, but that doesn't pass `self` to the
  `key` function, which we need to get the request URL. Long discussion:
  https://github.com/tkem/cachetools/issues/107

  Args:
    expiration: :class:`datetime.timedelta`
    size: integer, bytes. defaults to 20 MB.
  """
  lock = threading.RLock()
  ttlcache = cachetools.TTLCache(
    size, expiration.total_seconds(),
    getsizeof=lambda response: len(response.body))

  def decorator(method):
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
      cache = self.request.get('cache', '').lower() != 'false'
      if cache:
        resp = ttlcache.get(self.request.url)
        if resp:
          logging.info('Serving cached response')
          return resp

      resp = method(self, *args, **kwargs)
      if not resp:
        resp = self.response

      if cache:
        with lock:
          ttlcache[self.request.url] = resp

      return resp

    wrapper.cache_clear = ttlcache.clear
    return wrapper

  return decorator


def ndb_context_middleware(app, client=None):
  """WSGI middleware for per request instance info instrumentation.

  Follows the WSGI standard. Details: http://www.python.org/dev/peps/pep-0333/

  Install with e.g.:

    application = handlers.ndb_context_middleware(webapp2.WSGIApplication(...)

  Args:
    client: :class:`google.cloud.ndb.Client`
  """
  def wrapper(environ, start_response):
    with client.context():
      return app(environ, start_response)

  wrapper.get_response = app.get_response
  return wrapper


class ModernHandler(webapp2.RequestHandler):
  """Base handler that adds modern open/secure headers like CORS, HSTS, etc."""
  handle_exception = handle_exception

  def __init__(self, *args, **kwargs):
    super(ModernHandler, self).__init__(*args, **kwargs)
    self.response.headers.update({
      'Access-Control-Allow-Origin': '*',
      # see https://content-security-policy.com/
      'Content-Security-Policy':
        "script-src https: localhost:8080 my.dev.com:8080 'unsafe-inline'; "
        "frame-ancestors 'self'; "
        "report-uri /csp-report; ",
      # 16070400 seconds is 6 months
      'Strict-Transport-Security': 'max-age=16070400; preload',
      'X-Content-Type-Options': 'nosniff',
      # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Frame-Options
      'X-Frame-Options': 'SAMEORIGIN',
      'X-XSS-Protection': '1; mode=block',
    })


class TemplateHandler(ModernHandler):
  """Renders and serves a template based on class attributes.

  Subclasses must override :meth:`template_file()` and may also override
  :meth:`template_vars()` and :meth:`content_type()`.
  """
  def template_file(self):
    """Returns the string template file path."""
    raise NotImplementedError()

  def template_vars(self, *args, **kwargs):
    """Returns a dict of template variable string keys and values.

    Args:
      args, kwargs: passed through from get()
    """
    return {}

  def content_type(self):
    """Returns the string content type."""
    return 'text/html; charset=utf-8'

  def headers(self):
    """Returns dict of HTTP response headers. Subclasses may override.

    To advertise XRDS, use::

      headers['X-XRDS-Location'] = 'https://%s/.well-known/host-meta.xrds' % self.request.host
    """
    return {
      'Cache-Control': 'max-age=300',
      'Access-Control-Allow-Origin': '*',
      }

  def get(self, *args, **kwargs):
    self.response.headers['Content-Type'] = self.content_type()
    # can't update() because wsgiref.headers.Headers doesn't have it.
    for key, val in list(self.headers().items()):
      self.response.headers[key] = val

    vars = {
      'host': self.request.host,
      'host_uri': self.request.host_url,
    }

    # add query params. use a list for params with multiple values.
    for key in self.request.params:
      values = self.request.params.getall(key)
      if len(values) == 1:
        values = values[0]
      vars[key] = values

    vars.update(self.template_vars(*args, **kwargs))
    self.response.out.write(
      JINJA_ENV.get_template(self.template_file()).render(**vars))


class XrdOrJrdHandler(TemplateHandler):
  """Renders and serves an XRD or JRD file.

  JRD is served if the request path ends in .json, or the query parameters
  include 'format=json', or the request headers include
  'Accept: application/json'.

  Subclasses must override :meth:`template_prefix()`.

  Class members:
    JRD_TEMPLATE: boolean, renders JRD with a template if True,
    otherwise renders it as JSON directly.
  """
  JRD_TEMPLATE = True

  def get(self, *args, **kwargs):
    if self.JRD_TEMPLATE:
      return super(XrdOrJrdHandler, self).get(*args, **kwargs)

    self.response.headers['Content-Type'] = self.content_type()
    # can't update() because wsgiref.headers.Headers doesn't have it.
    for key, val in self.headers().items():
      self.response.headers[key] = val

    self.response.write(json_dumps(self.template_vars(*args, **kwargs), indent=2))

  def content_type(self):
    return ('application/json; charset=utf-8' if self.is_jrd()
            else 'application/xrd+xml; charset=utf-8')

  def template_prefix(self):
    """Returns template filename, without extension."""
    raise NotImplementedError()

  def template_file(self):
    return self.template_prefix() + ('.jrd' if self.is_jrd() else '.xrd')

  def is_jrd(self):
    """Returns True if JRD should be served, False if XRD."""
    return (os.path.splitext(self.request.path)[1] == '.json' or
            self.request.get('format') == 'json' or
            self.request.headers.get('Accept') == 'application/json')


class HostMetaHandler(XrdOrJrdHandler):
  """Renders and serves the /.well-known/host-meta file.
  """
  def template_prefix(self):
    return 'templates/host-meta'


class HostMetaXrdsHandler(TemplateHandler):
  """Renders and serves the /.well-known/host-meta.xrds XRDS-Simple file.
  """
  def content_type(self):
    return 'application/xrds+xml'

  def template_file(self):
    return 'templates/host-meta.xrds'


HOST_META_ROUTES = [
  ('/\.well-known/host-meta(?:\.json)?', HostMetaHandler),
  ('/\.well-known/host-meta.xrds', HostMetaXrdsHandler),
  ]



import logging, sys

# Piggyback on unittest's -v and -q flags to show/hide logging.
if 'discover' in sys.argv or '-q' in sys.argv or '--quiet' in sys.argv:
  logging.disable(logging.CRITICAL + 1)
elif '-v' in sys.argv:
  logging.getLogger().setLevel(logging.DEBUG)
  logging.getLogger('google.cloud').setLevel(logging.INFO)
  logging.getLogger('urllib3').setLevel(logging.INFO)

# Suppress warnings.
import warnings
# Not currently working.
# ../local3/lib/python3.6/site-packages/mox3/mox.py:909: DeprecationWarning: inspect.getargspec() is deprecated, use inspect.signature() or inspect.getfullargspec()
warnings.filterwarnings('ignore', module='mox', message='getargspec')



"""A handler that exposes App Engine app logs to users.

StackDriver Logging API:
https://cloud.google.com/logging/docs/apis
"""
import calendar
import cgi
import datetime
import logging
import re
import time
import urllib.request, urllib.parse, urllib.error

from google.cloud import ndb
from google.cloud.logging_v2 import LoggingServiceV2Client
import humanize
import webapp2

from .appengine_info import APP_ID
from . import handlers, util

LEVELS = {
  logging.DEBUG:    'D',
  logging.INFO:     'I',
  logging.WARNING:  'W',
  logging.ERROR:    'E',
  logging.CRITICAL: 'F',
}

CACHE_TIME = datetime.timedelta(days=1)
CACHE_SIZE = 2 * 1000 * 1000  # 2 MB

MAX_LOG_AGE = datetime.timedelta(days=30)
# App Engine's launch, roughly
MIN_START_TIME = time.mktime(datetime.datetime(2008, 4, 1).timetuple())

SANITIZE_RE = re.compile(r"""
  ((?:access|api|oauth)?[ _]?
   (?:code|consumer_key|consumer_secret|nonce|secret|signature|token|verifier)
     (?:u?['"])?
   (?:=|:|\ |,\ |%3D)\ *
     (?:u?['"])?
  )
  [^ &='"]+
""", flags=re.VERBOSE | re.IGNORECASE)

def sanitize(msg):
  """Sanitizes access tokens and Authorization headers."""
  return SANITIZE_RE.sub(r'\1...', msg)


def url(when, key):
  """Returns the relative URL (no scheme or host) to a log page.

  Args:
    when: datetime
    key: ndb.Key
  """
  return 'log?start_time=%s&key=%s' % (
    calendar.timegm(when.utctimetuple()), key.urlsafe().decode())


def maybe_link(when, key, time_class='dt-updated', link_class=''):
  """Returns an HTML snippet with a timestamp and maybe a log page link.

  Example:

  <a href="/log?start_time=1513904267&key=aglz..." class="u-bridgy-log">
    <time class="dt-updated" datetime="2017-12-22T00:57:47.222060"
            title="Fri Dec 22 00:57:47 2017">
      3 days ago
    </time>
  </a>

  The <a> tag is only included if the timestamp is 30 days old or less, since
  Stackdriver's basic tier doesn't store logs older than that:
    https://cloud.google.com/monitoring/accounts/tiers#logs_ingestion
    https://github.com/snarfed/bridgy/issues/767

  Args:
    when: datetime
    key: ndb.Key
    time_class: string, optional class value for the <time> tag
    link_class: string, optional class value for the <a> tag (if generated)

  Returns: string HTML
  """
  time = '<time class="%s" datetime="%s" title="%s">%s</time>' % (
    time_class, when.isoformat(), when.ctime(), humanize.naturaltime(when))

  now = datetime.datetime.now()
  if now > when > now - MAX_LOG_AGE:
    return '<a class="%s" href="/%s">%s</a>' % (link_class, url(when, key), time)

  return time


# datastore string keys are url-safe-base64 of, say, at least 32(ish) chars.
# https://cloud.google.com/appengine/docs/python/ndb/keyclass#Key_urlsafe
# http://tools.ietf.org/html/rfc3548.html#section-4
DATASTORE_KEY_RE = re.compile("'(([A-Za-z0-9-_=]{8})[A-Za-z0-9-_=]{24,})'")

def linkify_datastore_keys(msg):
  """Converts string datastore keys to links to the admin console viewer."""
  def linkify_key(match):
    try:
      key = ndb.Key(urlsafe=match.group(1))
      tokens = [(kind, '%s:%s' % ('id' if isinstance(id, int) else 'name', id))
                for kind, id in key.pairs()]
      key_str = '0/|' + '|'.join('%d/%s|%d/%s' % (len(kind), kind, len(id), id)
                                 for kind, id in tokens)
      key_quoted = urllib.parse.quote(urllib.parse.quote(key_str), safe=True)
      return "'<a title='%s' href='https://console.cloud.google.com/datastore/entities;kind=%s;ns=__$DEFAULT$__/edit;key=%s?project=%s'>%s...</a>'" % (
        match.group(1), key.kind(), key_quoted, APP_ID, match.group(2))
    except BaseException:
      logging.debug("Couldn't linkify candidate datastore key.", stack_info=True)
      return msg

  return DATASTORE_KEY_RE.sub(linkify_key, msg)


class LogHandler(webapp2.RequestHandler):
  """Searches for and renders the app logs for a single task queue request."""
  @handlers.cache_response(CACHE_TIME, size=CACHE_SIZE)
  def get(self):
    """URL parameters:
      start_time: float, seconds since the epoch
      key: string that should appear in the first app log
    """
    start_time = util.get_required_param(self, 'start_time')
    if not util.is_float(start_time):
      self.abort(400, "Couldn't convert start_time to float: %r" % start_time)

    start_time = float(start_time)
    if start_time < MIN_START_TIME:
      self.abort(400, "start_time must be >= %s" % MIN_START_TIME)

    logging_client = LoggingServiceV2Client()
    project = 'projects/%s' % APP_ID
    key = urllib.parse.unquote_plus(util.get_required_param(self, 'key'))

    # first, find the individual stdout log message to get the trace id
    query = '\
logName="%s/logs/stdout" AND \
timestamp>="%s" AND \
timestamp<="%s" AND \
jsonPayload.message:"%s"' % (
  project,
  datetime.datetime.utcfromtimestamp(start_time - 60).isoformat() + 'Z',
  datetime.datetime.utcfromtimestamp(start_time + 120).isoformat() + 'Z',
  key)
    logging.info('Searching logs with: %s', query)
    try:
      # https://googleapis.dev/python/logging/latest/gapic/v2/api.html#google.cloud.logginjg_v2.LoggingServiceV2Client.list_log_entries
      log = next(iter(logging_client.list_log_entries((project,), filter_=query, page_size=1)))
    except StopIteration:
      self.response.out.write('No log found!')
      return

    logging.info('Got insert id %s trace %s', log.insert_id, log.trace)

    # now, print all logs with that trace
    self.response.headers['Content-Type'] = 'text/html; charset=utf-8'
    self.response.out.write("""\
<html>
<body style="font-family: monospace; white-space: pre">
""")

    query = 'logName="%s/logs/stdout" AND trace="%s"' % (project, log.trace)
    logging.info('Searching logs with: %s', query)

    # sanitize and render each line
    for log in logging_client.list_log_entries((project,), filter_=query):
      msg = log.json_payload.fields['message'].string_value
      if msg:
        msg = linkify_datastore_keys(util.linkify(cgi.escape(
          msg if msg.startswith('Created by this poll:') else sanitize(msg))))
        timestamp = log.timestamp.seconds + float(log.timestamp.nanos) / 1000000000
        self.response.out.write('%s %s %s<br />' % (
          LEVELS[log.severity / 10],
          datetime.datetime.utcfromtimestamp(timestamp),
          msg.replace('\n', '<br />')))

    self.response.out.write('</body>\n</html>')



"""Unit tests for handlers.py.
"""
import datetime
import os
import socket
import traceback
import urllib.error

import webapp2

from .. import handlers
from ..testutil import HandlerTest

handlers.JINJA_ENV.loader.searchpath.append('/')


class FakeTemplateHandler(handlers.TemplateHandler):
  def template_file(self):
    return os.path.join(os.path.dirname(__file__), 'test_handler_get.tmpl')

  def template_vars(self):
    return {'foo': 'bar'}

  def content_type(self):
    return 'text/baz'

  def handle_exception(self, e, debug):
    traceback.print_exc()
    raise e


class HandlersTest(HandlerTest):

  def test_handle_exception(self):
    class Handler(webapp2.RequestHandler):
      handle_exception = handlers.handle_exception
      err = None
      def get(self):
        raise self.err

    app = webapp2.WSGIApplication([('/', Handler)])

    # HTTP exception
    Handler.err = urllib.error.HTTPError('/', 408, 'foo bar', None, None)
    resp = app.get_response('/')
    self.assertEqual(408, resp.status_int)
    self.assertEqual('HTTP Error 408: foo bar', resp.text)

    # network failure
    Handler.err = socket.timeout('foo bar')
    resp = app.get_response('/')
    self.assertEqual(504, resp.status_int)
    self.assertEqual('Upstream server request failed: foo bar', resp.text)

    # other exception
    Handler.err = AssertionError('foo')
    resp = app.get_response('/')
    self.assertEqual(500, resp.status_int)

  def test_redirect(self):
    class Handler(webapp2.RequestHandler):

      @handlers.redirect('from.com', 'to.org')
      def get(self):
        self.response.set_status(204)

      @handlers.redirect(('from.com', 'from.net'), 'to.org')
      def post(self, first, second):
        self.response.set_status(205)

    app = webapp2.WSGIApplication([('/(.*)/(.*)', Handler),
                                   ('.*', Handler)])

    # should redirect
    for url in '/', '/a/b/c', '/d?x=y':
      for scheme in 'http', 'https':
        resp = app.get_response(url, base_url=scheme + '://from.com')
        self.assertEqual(301, resp.status_int)
        self.assertEqual('%s://to.org%s' % (scheme, url), resp.headers['Location'])

    # should redirect and include *args
    for url in 'http://from.com', 'http://from.net':
      resp = app.get_response('/x/y', method='POST', base_url=url)
      self.assertEqual(301, resp.status_int)
      self.assertEqual('http://to.org/x/y', resp.headers['Location'])

    for base_url in 'http://abc.net', 'https://to.org':
      # shouldn't redirect
      resp = app.get_response('/', base_url=base_url)
      self.assertEqual(204, resp.status_int)
      self.assertNotIn('Location', resp.headers)

      # shouldn't redirect, should include *args
      resp = app.get_response('/x/y', method='POST', base_url='http://')
      self.assertEqual(205, resp.status_int)
      self.assertNotIn('Location', resp.headers)

  def test_template_handler_get_jinja(self):
    resp = webapp2.WSGIApplication([('/', FakeTemplateHandler)]).get_response('/')
    self.assertEqual("""\
my host: localhost:80
my foo: bar""", resp.text)

  def test_cache_response(self):
    class Handler(webapp2.RequestHandler):
      calls = 0

      @handlers.cache_response(datetime.timedelta(days=1))
      def get(self):
        Handler.calls += 1
        self.response.set_status(204)
        self.response.out.write('got %s' % self.request.url)

    app = webapp2.WSGIApplication([('.*', Handler)])

    # first fetch populates the cache
    resp = app.get_response('/?x')
    self.assertEqual(1, Handler.calls)
    self.assertEqual(204, resp.status_int)
    self.assertEqual('got http://localhost/?x', resp.text)

    # second fetch should use the cache instead of fetching from the silo
    resp = app.get_response('/?x')
    self.assertEqual(1, Handler.calls)
    self.assertEqual(204, resp.status_int)
    self.assertEqual('got http://localhost/?x', resp.text)

    # fetches with ?cache=false shouldn't use the cache
    resp = app.get_response('/?x&cache=false')
    resp = app.get_response('/?x&cache=false')
    self.assertEqual(3, Handler.calls)
    self.assertEqual(204, resp.status_int)
    self.assertEqual('got http://localhost/?x&cache=false', resp.text)

    # a different URL shouldn't be cached
    resp = app.get_response('/?y')
    self.assertEqual(4, Handler.calls)
    self.assertEqual(204, resp.status_int)
    self.assertEqual('got http://localhost/?y', resp.text)



"""HTTP request and server information derived from the environment.

App Engine: https://cloud.google.com/appengine/docs/standard/python3/runtime#environment_variables
WSGI: https://www.python.org/dev/peps/pep-3333/
"""
import os

APP_ID = os.getenv('GAE_APPLICATION', '').split('~')[-1]
DEBUG = os.environ.get('GAE_ENV') in (None, 'localdev')  # 'standard' in production
LOCAL = os.environ.get('GAE_ENV') == 'localdev'



"""Tumblr OAuth drop-in.

API docs:
http://www.tumblr.com/docs/en/api/v2
http://www.tumblr.com/oauth/apps
"""
import logging

from google.cloud import ndb
import tumblpy
from webob import exc

from . import handlers, models
from .webutil import handlers as webutil_handlers
from .webutil import util
from .webutil.util import json_dumps, json_loads

TUMBLR_APP_KEY = util.read('tumblr_app_key')
TUMBLR_APP_SECRET = util.read('tumblr_app_secret')


class TumblrAuth(models.BaseAuth):
  """An authenticated Tumblr user.

  Provides methods that return information about this user and make OAuth-signed
  requests to the Tumblr API. Stores OAuth credentials in the datastore. See
  models.BaseAuth for usage details.

  Tumblr-specific details: implements api() but not urlopen(). api()
  returns a tumblpy.Tumblpy. The datastore entity key name is the Tumblr
  username.
  """
  # access token
  token_key = ndb.StringProperty(required=True)
  token_secret = ndb.StringProperty(required=True)
  user_json = ndb.TextProperty(required=True)

  def site_name(self):
    return 'Tumblr'

  def user_display_name(self):
    """Returns the username.
    """
    return self.key_id()

  def access_token(self):
    """Returns the OAuth access token as a (string key, string secret) tuple.
    """
    return (self.token_key, self.token_secret)

  def _api(self):
    """Returns a tumblpy.Tumblpy.
    """
    return TumblrAuth._api_from_token(self.token_key, self.token_secret)

  @staticmethod
  def _api_from_token(key, secret):
    """Returns a tumblpy.Tumblpy.
    """
    assert TUMBLR_APP_KEY and TUMBLR_APP_SECRET, \
      "Please fill in the tumblr_app_key and tumblr_app_secret files in your app's root directory."
    return tumblpy.Tumblpy(app_key=TUMBLR_APP_KEY,
                           app_secret=TUMBLR_APP_SECRET,
                           oauth_token=key, oauth_token_secret=secret)


def handle_exception(self, e, debug):
  """Exception handler that handles Tweepy errors.
  """
  if isinstance(e, tumblpy.TumblpyError):
      logging.error('OAuth error', stack_info=True)
      raise exc.HTTPBadRequest(e)
  else:
    return webutil_handlers.handle_exception(self, e, debug)


class StartHandler(handlers.StartHandler):
  """Starts Tumblr auth. Requests an auth code and expects a redirect back.
  """
  NAME = 'tumblr'
  LABEL = 'Tumblr'

  handle_exception = handle_exception

  def redirect_url(self, state=None):
    assert TUMBLR_APP_KEY and TUMBLR_APP_SECRET, \
      "Please fill in the tumblr_app_key and tumblr_app_secret files in your app's root directory."
    tp = tumblpy.Tumblpy(app_key=TUMBLR_APP_KEY,
                         app_secret=TUMBLR_APP_SECRET)
    auth_props = tp.get_authentication_tokens(
      callback_url=self.request.host_url + self.to_path)

    # store the request token for later use in the callback handler
    models.OAuthRequestToken(id=auth_props['oauth_token'],
                             token_secret=auth_props['oauth_token_secret'],
                             state=state).put()
    return auth_props['auth_url']

  @classmethod
  def button_html(cls, *args, **kwargs):
    return super(cls, cls).button_html(
      *args,
      input_style='background-color: #406784; padding: 10px',
      **kwargs)


class CallbackHandler(handlers.CallbackHandler):
  """OAuth callback. Fetches the user's blogs and stores the credentials.
  """
  handle_exception = handle_exception

  def get(self):
    verifier = self.request.get('oauth_verifier')
    request_token_key = self.request.get('oauth_token')
    if not verifier or not request_token_key:
      # user declined
      self.finish(None)
      return

    # look up the request token
    request_token = models.OAuthRequestToken.get_by_id(request_token_key)
    if request_token is None:
      raise exc.HTTPBadRequest('Invalid oauth_token: %s' % request_token_key)

    # generate and store the final token
    tp = tumblpy.Tumblpy(app_key=TUMBLR_APP_KEY,
                         app_secret=TUMBLR_APP_SECRET,
                         oauth_token=request_token_key,
                         oauth_token_secret=request_token.token_secret)
    auth_token = tp.get_authorized_tokens(verifier)
    auth_token_key = auth_token['oauth_token']
    auth_token_secret = auth_token['oauth_token_secret']

    # get the user's blogs
    # http://www.tumblr.com/docs/en/api/v2#user-methods
    tp = TumblrAuth._api_from_token(auth_token_key, auth_token_secret)
    logging.debug('Fetching user/info')
    try:
      resp = tp.post('user/info')
    except BaseException as e:
      util.interpret_http_exception(e)
      raise
    logging.debug('Got: %s', resp)
    user = resp['user']

    auth = TumblrAuth(id=user['name'],
                      token_key=auth_token_key,
                      token_secret=auth_token_secret,
                      user_json=json_dumps(resp))
    auth.put()
    self.finish(auth, state=request_token.state)



"""setuptools setup module for oauth-dropins.

Docs:
https://packaging.python.org/en/latest/distributing.html
http://pythonhosted.org/setuptools/setuptools.html

Based on https://github.com/pypa/sampleproject/blob/master/setup.py
"""
from setuptools import setup, find_packages


setup(name='oauth-dropins',
      version='3.0',
      description='Drop-in App Engine OAuth client handlers for many popular sites.',
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      url='https://github.com/snarfed/oauth-dropins',
      packages=find_packages(),
      include_package_data = True,
      author='Ryan Barrett',
      author_email='oauth-dropins@ryanb.org',
      license='Public domain',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'Topic :: System :: Systems Administration :: Authentication/Directory',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Environment :: Web Environment',
          'License :: OSI Approved :: MIT License',
          'License :: Public Domain',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
      ],
      keywords='oauth appengine',
      install_requires=[
          'beautifulsoup4~=4.8',
          'cachetools~=3.1',
          'gdata-python3~=3.0',
          'google-cloud-ndb~=1.1',
          'humanize~=0.5',
          'jinja2~=2.10',
          'mf2py~=1.1,>=1.1.2',
          'mf2util>=0.5.0',
          'oauthlib~=3.1',
          'python-tumblpy~=1.1',
          'requests-oauthlib',
          'requests~=2.22',
          'tweepy~=3.7',
          'ujson~=1.35',
          'urllib3~=1.14',
          'webapp2>=3.0.0b1',
      ],
      tests_require=['mox3~=0.28'],
)
